#' @importFrom assertthat assert_that
#' @importFrom assertthat is.flag
#' @importFrom stats setNames update
#' @importFrom utils head tail
#' @importFrom glue glue
#' @importFrom cli cli_abort
#' @importFrom dplyr n
#' @import rlang
#' @import DBI
#' @importFrom tibble tibble as_tibble
#' @importFrom magrittr %>%
#' @keywords internal
"_PACKAGE"
